//-----------------------------------------------------------------------------
//
// Interface code for AmgX solver.
//
//-----------------------------------------------------------------------------

#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"
#include "cuda_runtime.h"
#include <amgx_c.h>

//-----------------------------------------------------------------------------

// CUDA safe call macro
#define CUDA_SAFE_CALL(call) do {                                 \
  cudaError_t err = call;                                         \
  if(cudaSuccess != err) {                                        \
    fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n", \
            __FILE__, __LINE__, cudaGetErrorString( err) );       \
    exit(EXIT_FAILURE);                                           \
  } } while (0)

// Assertion macro
//#define Assert(condition) {if ( !(condition) ) {                           \
//        fprintf(stderr, "Assertion error: \"%s\". At file %s, line %i.\n", \
//                #condition, __FILE__, __LINE__);                           \
//    exit(EXIT_FAILURE);                                                    \
//  }}
#define Assert(condition)

//-----------------------------------------------------------------------------
// AmgX solve options - see manual, examples for details.

const char* g_amgx_config_string =
  "{\n" 
  "    \"config_version\": 2, \n" 
  "    \"solver\": {\n" 
  "        \"preconditioner\": {\n" 
  "            \"print_grid_stats\": 0, \n" 
  "            \"print_vis_data\": 0, \n" 
  "            \"solver\": \"AMG\", \n" 
  "            \"algorithm\": \"CLASSICAL\",\n" 
  "            \"interpolator\": \"D2\",\n" 
  "            \"smoother\": {\n" 
  "                \"scope\": \"jacobi\", \n" 
  "                \"solver\": \"BLOCK_JACOBI\", \n" 
  "                \"monitor_residual\": 0, \n" 
  "                \"print_solve_stats\": 0\n" 
  "            }, \n" 
  "            \"print_solve_stats\": 0, \n" 
  "            \"presweeps\": 1, \n" 
  "            \"max_iters\": 1, \n" 
  "            \"monitor_residual\": 0, \n" 
  "            \"store_res_history\": 0, \n" 
  "            \"scope\": \"amg\", \n" 
  "            \"max_levels\": 100, \n" 
  "            \"cycle\": \"V\", \n" 
  "            \"postsweeps\": 1\n" 
  "        }, \n" 
  "        \"solver\": \"FGMRES\", \n" 
  "        \"gmres_n_restart\": 50,\n" //set to 50 instead of 20 for hard cases
  "        \"print_solve_stats\": 0, \n" 
  "        \"obtain_timings\": 0, \n" 
  "        \"max_iters\": 100, \n" 
  "        \"monitor_residual\": 1, \n" 
  "        \"convergence\": \"RELATIVE_INI_CORE\", \n" 
  "        \"scope\": \"main\", \n" 
  "        \"tolerance\": 1e-6,\n"
  "        \"norm\": \"L2\"\n" 
  "    }\n" 
  "}\n" ;

//-----------------------------------------------------------------------------

// Set relevant precisions.
// Always use double for linear solve, but data received from GTC
// may be single or double based on GTC build option.

typedef double Float_t;

#ifdef DOUBLE_PRECISION
typedef double GTC_float_t;
#else
typedef float GTC_float_t;
#endif

// Struct to hold GTC/AmgX solver state.

struct AmgxCommon {
  AMGX_config_handle cfg;
  AMGX_resources_handle rsrc;
  AMGX_vector_handle x;
  AMGX_vector_handle b;
  AMGX_matrix_handle a;
  AMGX_solver_handle solver;
  MPI_Comm comm;
  int comm_size;
  int comm_rank;
  Float_t* h_x;
  Float_t* h_x_gathered;
  Float_t* h_b;
  Float_t* d_x;
  Float_t* d_b;
  int* recvcounts;
  int* displs;
  int n_global;
  int ilo;
  int ihi;
  int n_local;
};

// Global variable to hold solve state info,
// analogous to GTC Fortran common block.

AmgxCommon g_amgx_common2;

//-----------------------------------------------------------------------------
// Helper functions.

enum {AMGX_VALUE_UNSPECIFIED = 0};

int amgx_ranks_per_solve() {
  char* env_ranks_per_solve =  getenv("GTC_RANKS_PER_SOLVE");
  int result = env_ranks_per_solve ? atoi(env_ranks_per_solve)
                                   : AMGX_VALUE_UNSPECIFIED;
  result = result > 0 ? result : AMGX_VALUE_UNSPECIFIED;
  return result;
}

int amgx_solves_per_grid() {
  char* env_solves_per_grid =  getenv("GTC_SOLVES_PER_GRID");
  int result = env_solves_per_grid ? atoi(env_solves_per_grid)
                                   : AMGX_VALUE_UNSPECIFIED;
  result = result > 0 ? result : AMGX_VALUE_UNSPECIFIED;
  return result;
}

//-----------------------------------------------------------------------------
// Get base index in global coords for part of vector on this rank.

int amgx_vec_index(AmgxCommon* data, int proc) {
  return (data->n_global * (size_t)proc) / data->comm_size;
}

//-----------------------------------------------------------------------------
// AmgX basic initializations.

extern "C"
void amgx_gtc_initialize_(int* partd_comm_fortran) {

  // Select struct to use.

  AmgxCommon* const data = &g_amgx_common2;

  // Initialize AmgX.

  AMGX_SAFE_CALL(AMGX_initialize());
  AMGX_SAFE_CALL(AMGX_initialize_plugins());

  //int major, minor;
  //AMGX_get_api_version(&major, &minor);
  //printf("AMGX_get_api_version result: %i %i.\n", major, minor);

  // Get the communicator to run a solve in.
  // Must convert from Fortran communicator to C communicator.

  MPI_Comm partd_comm = MPI_Comm_f2c(*partd_comm_fortran);

  // Determine how to parallelize the solves.

  int ranks_per_grid = 0;
  MPI_Comm_size(partd_comm, &ranks_per_grid);

  // TODO: better error checking - notify user if invalid value given.

  int solves_per_grid_requested = amgx_solves_per_grid();

  int ranks_per_solve = solves_per_grid_requested != AMGX_VALUE_UNSPECIFIED ?
      ranks_per_grid / solves_per_grid_requested : amgx_ranks_per_solve();
  ranks_per_solve = ranks_per_solve > 0 ? ranks_per_solve
                                        : AMGX_VALUE_UNSPECIFIED;

  if (ranks_per_solve == AMGX_VALUE_UNSPECIFIED) {
    // Default: distribute solve across all available ranks.
    MPI_Comm_dup(partd_comm, &data->comm);
  } else {
    // Distribute each solve across a requested number of ranks.
    int partd_comm_rank;
    MPI_Comm_rank(partd_comm, &partd_comm_rank);

    const int color = partd_comm_rank / ranks_per_solve;
    const int key = partd_comm_rank % ranks_per_solve;

    MPI_Comm_split(partd_comm, color, key, &data->comm);
  }

  MPI_Comm_size(data->comm, &data->comm_size);
  MPI_Comm_rank(data->comm, &data->comm_rank);
}

//-----------------------------------------------------------------------------
// Terminations - undo effects of amgx_initialize_, amgx_create_.

extern "C"
void amgx_gtc_finalize_() {

  // Select struct to use.

  AmgxCommon* const data = &g_amgx_common2;  

  free(data->recvcounts);
  free(data->displs);

  CUDA_SAFE_CALL(cudaFreeHost(data->h_b));
  CUDA_SAFE_CALL(cudaFreeHost(data->h_x));
  CUDA_SAFE_CALL(cudaFreeHost(data->h_x_gathered));

  CUDA_SAFE_CALL(cudaFree(data->d_b));
  CUDA_SAFE_CALL(cudaFree(data->d_x));

  AMGX_solver_destroy(data->solver);
  AMGX_matrix_destroy(data->a);
  AMGX_vector_destroy(data->b);
  AMGX_vector_destroy(data->x);

  AMGX_resources_destroy(data->rsrc);
  AMGX_SAFE_CALL(AMGX_config_destroy(data->cfg));

  MPI_Comm_free(&data->comm);

  AMGX_SAFE_CALL(AMGX_finalize_plugins());
  AMGX_SAFE_CALL(AMGX_finalize());
}

//-----------------------------------------------------------------------------
// Set up vectors, matrix and solver.

extern "C"
void amgx_gtc_create_(int* n_global, int* lusers, int* luserp,
                      GTC_float_t* lusera) {

  // Select struct to use.

  AmgxCommon* const data = &g_amgx_common2;

  // The GTC partd communicator corresponds to the grid for a single
  // 2D slice of the torus.
  // Every MPI rank in this communicator contains a replicated copy
  // of the entire linear system for this communicator.
  // To parallelize this within the communicator, a block row
  // decomposition of the system is imposed, and each rank takes
  // its part of the linear system from the global system to do the solve.
  // At the end of the solve, each rank has part of the solution.
  // An MPI allgatherv is performed to give every rank the entire
  // solution vector, as GTC needs.

  //---
  // Initialize general AmgX settings.
  //---

  //Could read from file if desired:
  //const char* cfg__filename = "amgx_config.json";
  //AMGX_SAFE_CALL(AMGX_config_create_from_file(&data->cfg, cfg__filename));
  AMGX_SAFE_CALL(AMGX_config_create(&data->cfg, g_amgx_config_string));
  AMGX_SAFE_CALL(AMGX_config_add_parameters(&data->cfg,
                                            "exception_handling=1"));
  // Use single GPU per MPI rank.
  int igpu;
  CUDA_SAFE_CALL(cudaGetDevice(&igpu));
  int devices[] = {igpu};
  AMGX_resources_create(&data->rsrc, data->cfg, &data->comm, 1, devices);

  const AMGX_Mode mode = AMGX_mode_dDDI;

  //---
  // Set up vectors.
  //---

  // Get global and local vector sizes and local vector start/end coords.

  // NOTE: 32 bit integers being used here, may be a problem in the future.

  data->n_global = *n_global;

  data->ilo = amgx_vec_index(data, data->comm_rank);
  data->ihi = amgx_vec_index(data, data->comm_rank+1);

  data->n_local = data->ihi - data->ilo;

  // Allocate vectors: AmgX vectors, CPU and GPU vector data arrays.

  AMGX_vector_create(&data->b, data->rsrc, mode);
  AMGX_vector_create(&data->x, data->rsrc, mode);

  CUDA_SAFE_CALL(cudaMalloc((void**)&data->d_b, data->n_local*sizeof(Float_t)));
  CUDA_SAFE_CALL(cudaMalloc((void**)&data->d_x, data->n_local*sizeof(Float_t)));

  CUDA_SAFE_CALL(cudaMallocHost((void**)&data->h_b,
                                data->n_local*sizeof(Float_t)));
  CUDA_SAFE_CALL(cudaMallocHost((void**)&data->h_x,
                                data->n_local*sizeof(Float_t)));
  CUDA_SAFE_CALL(cudaMallocHost((void**)&data->h_x_gathered,
                                data->n_global*sizeof(Float_t)));

  // Set up arrays needed for final vector gather.

  data->recvcounts = (int*)malloc(data->comm_size*sizeof(*(data->recvcounts)));
  data->displs = (int*)malloc(data->comm_size*sizeof(*(data->displs)));

  for (int p=0; p<data->comm_size; ++p) {
    const int ilo = amgx_vec_index(data, p);
    const int ihi = amgx_vec_index(data, p+1);
    data->recvcounts[p] = ihi - ilo;
    data->displs[p] = ilo;
  }

  //---
  // Set up matrix.
  //---


  const int nnz_local = lusers[data->ihi] - lusers[data->ilo];

  int* row_ptrs = (int*)malloc((data->n_local+1)*sizeof(*row_ptrs));
  int64_t* col_indices_global = (int64_t*)malloc(nnz_local*
                                                 sizeof(*col_indices_global));
  Float_t* values = (Float_t*)malloc(nnz_local*sizeof(*values));

  // Create required sparse matrix storage arrays.

  int i0 = 0;
  int j0 = 0;
  row_ptrs[i0] = 0;
  for (int i=data->ilo; i<data->ihi; ++i) {
    for (int j=lusers[i]; j<lusers[i+1]; ++j) {
      Assert(j0 < nnz_local);
      col_indices_global[j0] = luserp[j];
      values[j0] = (Float_t)(lusera[j]);
      Assert(col_indices_global[j0] >= 0);
      Assert(col_indices_global[j0] < data->n_global);
      ++j0;
    } //---j
    Assert(i0+1 < data->n_local+1);
    row_ptrs[i0+1] = row_ptrs[i0] + lusers[i+1] - lusers[i];
    Assert(row_ptrs[i0+1] >= row_ptrs[i0]);
    Assert(row_ptrs[i0+1] <= nnz_local);
    ++i0;
  } //---i
  Assert(row_ptrs[data->n_local] == nnz_local);

  // AmgX requires a partitioning vector to indicate mapping of rows to ranks.

  int* partition_vector = (int*)malloc(data->n_global*
                                       sizeof(*partition_vector));

  for (int p=0; p<data->comm_size; ++p) {
    const int ilo = amgx_vec_index(data, p);
    const int ihi = amgx_vec_index(data, p+1);
    for (int i=ilo; i<ihi; ++i) {
      Assert(i>=0);
      Assert(i<data->n_global);
      partition_vector[i] = p;
    } //---i
  } //---p

  // Allocate and set matrix: AmgX matrix.

  AMGX_matrix_create(&data->a, data->rsrc, mode);

  int nrings; // = 2 for classical R/S AMG;
  AMGX_config_get_default_number_of_rings(data->cfg, &nrings);
  //printf("nrings=%d\n",nrings);

  AMGX_matrix_upload_all_global(
    data->a, data->n_global, data->n_local, nnz_local,
    1, 1, row_ptrs, col_indices_global, values,
    NULL, nrings, nrings, partition_vector);

  AMGX_vector_bind(data->b, data->a);
  AMGX_vector_bind(data->x, data->a);

  free(row_ptrs);
  free(col_indices_global);
  free(values);
  free(partition_vector);

  //---
  // Set up solver.
  //---

  AMGX_solver_create(&data->solver, data->rsrc, mode, data->cfg);
  AMGX_solver_setup(data->solver, data->a);
}

//-----------------------------------------------------------------------------
// Perform a single linear system solve.

extern "C"
void amgx_gtc_solver_(GTC_float_t* luserx, GTC_float_t* luserb) {

  // Select struct to use.

  AmgxCommon* const data = &g_amgx_common2;  

  // Extract portion of global rhs, initial guess to use for solve on this rank.

  GTC_float_t* h_b = luserb + data->ilo;
  GTC_float_t* h_x = luserx + data->ilo;

  for (int i=0; i<data->n_local; ++i) {
    data->h_b[i] = (Float_t)(h_b[i]);
    data->h_x[i] = (Float_t)(h_x[i]);
  }

  // Copy vectors to GPU.

  CUDA_SAFE_CALL(cudaMemcpy(data->d_b, data->h_b, data->n_local*sizeof(Float_t),
                            cudaMemcpyHostToDevice));

  CUDA_SAFE_CALL(cudaMemcpy(data->d_x, data->h_x, data->n_local*sizeof(Float_t),
                            cudaMemcpyHostToDevice));

  // "Upload" GPU arrays to AmgX vector objects.

  AMGX_vector_upload(data->b, data->n_local, 1, data->d_b);
  AMGX_vector_upload(data->x, data->n_local, 1, data->d_x);

  // Perform solve.

  AMGX_solver_solve(data->solver, data->b, data->x);

  // "Download" solution vector from AmgX vector object to GPU array.

  AMGX_vector_download(data->x, data->d_x);

  // Copy GPU array to CPU.

  CUDA_SAFE_CALL(cudaMemcpy(data->h_x, data->d_x, data->n_local*sizeof(Float_t),
                            cudaMemcpyDeviceToHost));

  // Distribute each rank's portion of solution to all ranks.

  MPI_Allgatherv(data->h_x, data->n_local,
                 sizeof(Float_t) == 8 ? MPI_DOUBLE : MPI_FLOAT,
                 data->h_x_gathered, data->recvcounts, data->displs,
                 sizeof(Float_t) == 8 ? MPI_DOUBLE : MPI_FLOAT, data->comm);

  for (int i=0; i<data->n_global; ++i) {
    luserx[i] = (GTC_float_t)(data->h_x_gathered[i]);
  }
}

//-----------------------------------------------------------------------------
